import unittest

import lxml.etree

from congress_crawler import committee_meetings, utils


# Parsing the House hearing info


class HearingInfo(unittest.TestCase):
    def test_hearing(self):
        committees = {}
        for c in utils.yaml_load("test/fixtures/committees-current.yaml"):
            committees[c["thomas_id"]] = c
            if "house_committee_id" in c:
                committees[c["house_committee_id"] + "00"] = c
            c["subcommittees"] = {s["thomas_id"]: s for s in c.get("subcommittees", [])}

        hearing_xml = "test/fixtures/hearings/sample_hearing.xml"
        file_xml = open(hearing_xml)
        dom = lxml.etree.parse(file_xml)
        test_output = committee_meetings.parse_house_committee_meeting(
            "102252",
            dom,
            [],
            committees,
            {"debug": False},
            None,
            ["BILLS-113hr4435ih.pdf", "BILLS-113hr4435ih.xml"],
        )

        #          event_id, dom, existing_meetings, committees, options, witnesses, uploaded_documents
        self.assertEqual(test_output["bill_ids"], ["hr4435-113"])
        self.assertEqual(test_output["chamber"], "house")
        self.assertEqual(test_output["committee"], "HSRU")
        self.assertEqual(test_output["congress"], 113)
        self.assertEqual(test_output["house_meeting_type"], "HMTG")
        self.assertEqual(
            test_output["meeting_documents"][0]["description"],
            "H.R. 4435 (as introduced)",
        )
        self.assertEqual(test_output["meeting_documents"][0]["bill_id"], "hr4435-113")
        self.assertEqual(test_output["meeting_documents"][0]["version_code"], "ih")
        self.assertEqual(test_output["meeting_documents"][0]["type"], "BR")
        self.assertEqual(
            test_output["meeting_documents"][0]["urls"],
            [
                {
                    "url": "http://beta.congress.gov/113/bills/hr4435/BILLS-113hr4435ih.pdf",
                    "file_found": True,
                },
                {
                    "url": "http://beta.congress.gov/113/bills/hr4435/BILLS-113hr4435ih.xml",
                    "file_found": True,
                },
            ],
        )
        self.assertEqual(test_output["occurs_at"], "2014-05-19T17:00:00")
        self.assertEqual(test_output["room"], "CAPITOL H-313")
        self.assertEqual(test_output["subcommittee"], None)
        self.assertEqual(
            test_output["topic"],
            "H.R. 4435\u2014National Defense Authorization Act for Fiscal Year 2015 [General Debate]; H.R. 4660\u2014Commerce, Justice, Science, and Related Agencies Appropriations Act, 2015",
        )
        self.assertEqual(
            test_output["url"],
            "http://docs.house.gov/Committee/Calendar/ByEvent.aspx?EventID=102252",
        )

    def test_witnesses(self):
        witness_xml = "test/fixtures/hearings/sample_witness.xml"
        file_xml = open(witness_xml)
        witness_tree = lxml.etree.parse(file_xml)
        uploaded_documents = ["HHRG-113-GO25-Bio-CochraneJ-20140522.pdf"]
        event_id = "102266"
        test_output = committee_meetings.parse_witness_list(
            witness_tree, uploaded_documents, event_id
        )["hearing_witness_info"]

        self.assertEqual(test_output[0]["documents"][0]["type"], "WB")
        self.assertEqual(test_output[0]["documents"][0]["description"], "Cochrane Bio")
        self.assertEqual(
            test_output[0]["documents"][0]["urls"],
            [
                {
                    "url": "http://docs.house.gov/meetings/GO/GO25/20140522/102266/HHRG-113-GO25-Bio-CochraneJ-20140522.pdf",
                    "file_found": True,
                }
            ],
        )
        self.assertEqual(test_output[0]["house_event_id"], "102266")
        self.assertEqual(test_output[0]["first_name"], "James")
        self.assertEqual(test_output[0]["last_name"], "Cochrane")
        self.assertEqual(
            test_output[0]["position"],
            "Chief Information Officer and Executive Vice President",
        )
        self.assertEqual(test_output[0]["organization"], "U.S. Postal Service")



# Import the Adler & Wilkerson Congressional Bills Project
# data, which covers bills (but not resolutions) in the
# 80th through 92nd Congress,

import csv
import datetime
import zipfile

from congress_crawler import utils


def run(options):
    # Download the TSV file.
    cache_zip_path = "adler-wilkerson-bills.zip"
    utils.download(
        "http://congressionalbills.org/billfiles/bills80-92.zip",
        cache_zip_path,
        utils.merge(options, {"binary": True, "needs_content": False}),
    )

    # Unzip in memory and process the records.
    zfile = zipfile.ZipFile(utils.cache_dir() + "/" + cache_zip_path)
    csvreader = csv.DictReader(zfile.open("bills80-92.txt"), delimiter="\t")
    for record in csvreader:
        rec = process_bill(record)

        import pprint

        pprint.pprint(rec)


def process_bill(record):
    # Basic info
    congress = int(record["Cong"])
    bill_type = record["BillType"].lower()  # "HR" or "S" only
    if bill_type not in ("hr", "s"):
        raise ValueError(bill_type)
    number = int(record["BillNum"])

    def binary(value):
        if value == "NULL":
            return None
        return value == "1"

    def nullydate(value):
        if value == "NULL":
            return None
        raise ValueError(value)  # never occurs -- there are no dates in the dataset!

    # Last status?
    status = "INTRODUCED"
    status_at = nullydate(record["IntrDate"])
    if record["ReportH"] == "1" or record["ReportS"] == "1":
        status = "REPORTED"
    if record["PassH"] == "1" and record["PassS"] == "1":
        status = "PASSED:BILL"
    elif record["PassH"] == "1":
        status = "PASS_OVER:HOUSE"
    elif record["PassS"] == "1":
        status = "PASS_OVER:SENATE"
    if record["PLaw"] == "1":
        if record["Veto"] == "1":
            status = "ENACTED:VETO_OVERRIDE"
        else:
            status = "ENACTED:SIGNED"  # could also have been a 10-day rule
        status_at = nullydate(record["PLawDate"])
    else:
        if record["Veto"] == "1":
            status = "PROV_KILL:VETO"

    # Form data structure
    return {
        "bill_id": "%s%d-%d" % (bill_type, number, congress),
        "bill_type": bill_type,
        "number": number,
        "congress": congress,
        "introduced_at": nullydate(record["IntrDate"]),
        "sponsor": int(record["PooleID"]) if record["PooleID"] != "NULL" else None,
        #'cosponsors': ,
        #'actions': ,
        "history": {
            "house_passage_result": "pass" if record["PassH"] == "1" else None,
            "senate_passage_result": "pass" if record["PassS"] == "1" else None,
            "enacted": record["PLaw"] == "1",
            "enacted_at": nullydate(record["PLawDate"]),
        },
        "status": status,
        "status_at": status_at,
        "enacted_as": {
            "law_type": "public",
            "congress": congress,
            "number": int(record["PLawNum"]),
        }
        if record["PLaw"] == "1"
        else None,  # private laws?
        #'titles': ,
        "official_title": record["Title"],
        #'short_title': ,
        #'popular_title': ,
        #'summary': ,
        "subjects_top_term": int(record["Major"]),
        "subjects": [int(record["Minor"])],
        #'related_bills': ,
        #'committees': ,
        #'amendments': ,
        # special fields
        "by_request": binary(record["ByReq"]),
        "commemerative": binary(record["Commem"]),
        "num_cosponsors": int(record["Cosponsr"])
        if record["Cosponsr"] != "NULL"
        else None,
        "private": binary(record["Private"]),
        # meta-metadata
        "updated_at": datetime.datetime.now(),
    }



import json
import logging
import re
from datetime import datetime

from lxml.html import fromstring

from congress_crawler import utils


# can be run on its own, just require a nomination_id (e.g. PN2094-112)


def run(options):
    nomination_id = options.get("nomination_id", None)

    if nomination_id:
        result = fetch_nomination(nomination_id, options)
        logging.warning("\n%s" % result)
    else:
        logging.error("To run this task directly, supply a bill_id.")


# download and cache page for nomination


def fetch_nomination(nomination_id, options={}):
    logging.info("\n[%s] Fetching..." % nomination_id)

    # fetch committee name map, if it doesn't already exist
    nomination_type, number, congress = utils.split_nomination_id(nomination_id)
    if not number:
        return {
            "saved": False,
            "ok": False,
            "reason": "Couldn't parse %s" % nomination_id,
        }

    if not utils.committee_names:
        utils.fetch_committee_names(congress, options)

    # fetch bill details body
    body = utils.download(
        nomination_url_for(nomination_id),
        nomination_cache_for(nomination_id, "information.html"),
        options,
    )

    if not body:
        return {"saved": False, "ok": False, "reason": "failed to download"}

    if options.get("download_only", False):
        return {"saved": False, "ok": True, "reason": "requested download only"}

    # TODO:
    #   detect group nominations, particularly for military promotions
    #   detect when a group nomination is split into subnominations
    #
    # Also, the splitting process is nonsense:
    # http://thomas.loc.gov/home/PN/split.htm

    if "split into two or more parts" in body:
        return {"saved": False, "ok": True, "reason": "was split"}

    nomination = parse_nomination(nomination_id, body, options)
    output_nomination(nomination, options)
    return {"ok": True, "saved": True}


def parse_nomination(nomination_id, body, options):
    nomination_type, number, congress = utils.split_nomination_id(nomination_id)

    # remove (and store) comments, which contain some info for the nomination
    # but also mess up the parser
    facts = re.findall("<!--(.+?)-->", body)
    body = re.sub("<!--.+?-->", "", body)

    doc = fromstring(body)

    # get rid of centered bold labels, they screw stuff up,
    # e.g. agency names on PN1375-113
    body = re.sub(re.compile('<div align="center">.+?</div>', re.M), "", body)
    for elem in doc.xpath('//div[@align="center"]'):
        elem.getparent().remove(elem)

    committee_names = []
    committees = []

    info = {"nomination_id": nomination_id, "actions": []}

    # the markup on these pages is a disaster, so we're going to use a heuristic based on boldface, inline tags followed by text
    for pair in doc.xpath('//span[@class="elabel"]|//strong'):
        if pair.tail:
            text = pair.text or pair.text_content()
            label, data = text.replace(":", "").strip(), pair.tail.strip()

            # handle actions separately
            if label.split(" ")[-1] == "Action":
                pieces = re.split(r"\s+\-\s+", data)

                location = label.split(" ")[0].lower()

                # use 'acted_at', even though it's always a date, to be consistent
                # with acted_at field on bills and amendments
                acted_at = datetime.strptime(pieces[0], "%B %d, %Y").strftime(
                    "%Y-%m-%d"
                )

                # join rest back together (in case action itself has a hyphen)
                text = str.join(" - ", pieces[1 : len(pieces)])

                info["actions"].append(
                    {
                        "type": "action",
                        "location": location,
                        "acted_at": acted_at,
                        "text": text,
                    }
                )

            else:
                # let's handle these cases one by one
                if label == "Organization":
                    info["organization"] = data

                elif label == "Control Number":
                    # this doesn't seem useful
                    pass

                elif label.lower() == "referred to":
                    committee_names.append(data)

                elif label == "Reported by":
                    info["reported_by"] = data

                elif label == "Nomination":
                    # sanity check - verify nomination_id matches
                    if nomination_id != data:
                        raise Exception("Whoa! Mismatched nomination ID.")

                elif label == "Date Received":
                    # Note: Will break with the 1000th congress in year 3789
                    match = re.search(r"(\d{2,3})[stndhr]{2}", data)
                    if match:
                        info["congress"] = int(match.group(1))
                    else:
                        raise Exception(
                            'Choked, couldn\'t find Congress in "%s"' % data
                        )

                    # Doc format is: "January 04, 1995 (104th Congress)"
                    info["received_on"] = datetime.strptime(
                        data.split(" (")[0], "%B %d, %Y"
                    ).strftime("%Y-%m-%d")

                elif label == "Nominee":

                    # ignore any vice suffix
                    name = data.split(", vice")[0]

                    try:
                        name = re.search("(.+?),", name).groups()[0]
                    except Exception as e:
                        raise Exception("Couldn't parse nominee entry: %s" % name)

                    # Some begin "One nomination,...", so 'List of Nominees' will get it
                    if "nomination" in name:
                        pass

                    # and grab the state and position out of the comment facts
                    if facts[-5]:
                        position = facts[-5]
                    else:
                        raise Exception("Couldn't find the position in the comments.")

                    info["nominees"] = [
                        {"name": name, "position": position, "state": facts[-6][2:]}
                    ]

                elif label.lower() == "nominees":
                    pass

                elif label.lower() == "authority date":
                    pass

                elif label.lower() == "list of nominees":
                    # step through each sibling, collecting each br's stripped tail for names as we go
                    # stop when we get to a strong or span (next label)
                    nominees = []

                    current_position = None
                    for sibling in pair.itersiblings():
                        if sibling.tag == "br":
                            if sibling.tail:
                                name = sibling.tail.strip()
                                if name[0:5].lower() == "to be":
                                    current_position = name[6:].strip()
                                elif name:
                                    nominees.append(
                                        {
                                            "name": sibling.tail.strip(),
                                            "position": current_position,
                                        }
                                    )
                        elif (sibling.tag == "strong") or (sibling.tag == "span"):
                            break

                    info["nominees"] = nominees

                else:
                    # choke, I think we handle all of them now
                    raise Exception("Unrecognized label: %s" % label)

    if not info.get("received_on", None):
        raise Exception("Choked, couldn't find received date.")

    if not info.get("nominees", None):
        raise Exception("Choked, couldn't find nominee info.")

    # try to normalize committee name to an ID
    # choke if it doesn't work - the names should match up.
    for committee_name in committee_names:
        committee_id = utils.committee_names[committee_name]
        committees.append(committee_id)
    info["referred_to"] = committees
    info["referred_to_names"] = committee_names

    return info


# directory helpers


def output_for_nomination(nomination_id, format):
    nomination_type, number, congress = utils.split_nomination_id(nomination_id)
    return "{}/{}/nominations/{}/{}".format(
        utils.data_dir(), congress, number, "data.%s" % format
    )


def nomination_url_for(nomination_id):
    nomination_type, number, congress = utils.split_nomination_id(nomination_id)

    # numbers can be either of the form "63" or "64-01"
    number_pieces = number.split("-")
    if len(number_pieces) == 1:
        number_pieces.append("00")
    url_number = "%05d%s" % (int(number_pieces[0]), number_pieces[1])

    return "http://thomas.loc.gov/cgi-bin/ntquery/z?nomis:%03d%s%s:/" % (
        int(congress),
        nomination_type.upper(),
        url_number,
    )


def nomination_cache_for(nomination_id, file):
    nomination_type, number, congress = utils.split_nomination_id(nomination_id)
    return f"{congress}/nominations/{number}/{file}"


def output_nomination(nomination, options):
    logging.info("[%s] Writing to disk..." % nomination["nomination_id"])

    # output JSON - so easy!
    utils.write(
        json.dumps(nomination, sort_keys=True, indent=2, default=utils.format_datetime),
        output_for_nomination(nomination["nomination_id"], "json"),
    )



import datetime
import unittest

from congress_crawler.amendment_info import parse_amendment_actions


# parsing various kinds of action text to extract metadata and establish state


class AmendmentActions(unittest.TestCase):
    def test_amendment_action_uses_amendments_plural(self):
        action = {
            "text": "On agreeing to the Poe amendments (A009) Failed by recorded vote: 141 - 279 (Roll no. 164).",
            "type": "action",
            "references": [],
            "acted_at": datetime.datetime(2005, 6, 17, 11, 16),
        }
        actions = [action]
        parse_amendment_actions(actions)
        action = actions[0]
        self.assertEqual(action["where"], "h")
        self.assertEqual(action["type"], "vote")
        self.assertEqual(action["vote_type"], "vote")
        self.assertEqual(action["result"], "fail")
        self.assertEqual(action["how"], "roll")
        self.assertEqual(action["roll"], 164)

    def test_amendment_action_uses_as_modified_instead_of_as_amended(self):
        action = {
            "text": "On agreeing to the Jackson-Lee (TX) amendment (A015) as modified Agreed to by recorded vote: 233 - 192 (Roll no. 412). (text as modified: CR H6290)",
            "type": "action",
            "references": [],
            "acted_at": datetime.datetime(2005, 6, 17, 11, 16),
        }
        actions = [action]
        parse_amendment_actions(actions)
        action = actions[0]
        self.assertEqual(action["where"], "h")
        self.assertEqual(action["type"], "vote")
        self.assertEqual(action["vote_type"], "vote")
        self.assertEqual(action["result"], "pass")
        self.assertEqual(action["how"], "roll")
        self.assertEqual(action["roll"], 412)

    def test_amendment_action_uses_capital_N_in_no(self):
        action = {
            "text": "On agreeing to the Capps amendment (A028) Failed by recorded vote: 213 - 219 (Roll No. 129).",
            "type": "action",
            "references": [],
            "acted_at": datetime.datetime(2005, 6, 17, 11, 16),
        }
        actions = [action]
        parse_amendment_actions(actions)
        action = actions[0]
        self.assertEqual(action["where"], "h")
        self.assertEqual(action["type"], "vote")
        self.assertEqual(action["vote_type"], "vote")
        self.assertEqual(action["result"], "fail")
        self.assertEqual(action["how"], "roll")
        self.assertEqual(action["roll"], 129)



# Convert GPO Fdsys STATUTE metadata into bill files.
#
# GPO has the Statutes at Large from 1951 (the 65th
# volume, 82nd Congress) to the present, with metadata
# at the level of the law.
#
# The bill files have sort of made up action entries
# since we don't know the legislative history of the bill.
# We also assume all bills are enacted by being signed
# by the President for the sake of outputting status
# information.
#
# First download the Statutes at Large from GPO:
#
# ./run fdsys --collections=STATUTE --store=mods
#
# To process statute text, get the text PDFs:
#
# ./run fdsys --collections=STATUTE --store=pdfs --granules
#
# Then run this script:
#
# ./run statutes
#
# Processes all downloaded statutes files and saves bill files:
#   data/82/bills/hr/hr1/data.json and
#   data/82/bills/hr/hr1/text-versions/enr/data.json
#
# Specify --textversions to only write the text-versions file.
#
# If the individual statute PDF files are available, then
# additional options are possible:
#
# If --linkpdf is given, then *hard links* are created from
# where the PDF should be for bill text to where the PDF has
# been downloaded in the fdsys directory.
#
# If --extracttext is given, then the pdf is converted to text
# using "pdftotext -layout" and they are stored in files like
# data/82/bills/hr/hr1/text-versions/enr/document.txt. They are
# UTF-8 encoded and have form-feed characters marking page breaks.
#
# Examples:
# ./run statutes --volume=65
# ./run statutes --volumes=65-86
# ./run statutes --year=1951
# ./run statutes --years=1951-1972
# Processes just the indicated volume or range of volumes.
# Starting with the 93rd Congress (1973-1974, corresponding
# to volume 78 of the Statutes of Large), we have bill
# data from THOMAS. Be careful not to overwrite those files.
#
# With bill text missing from THOMAS/GPO from the 93rd to
# 102nd Congresses, fill in the text-versions files like so:
# ./run statutes --volumes=87-106 --textversions

import datetime
import glob
import json
import logging
import os.path
import subprocess
import time

from lxml import etree

from congress_crawler import bill_info, bills, govinfo, utils


def run(options):
    root_dir = utils.data_dir() + "/fdsys/STATUTE"

    if "volume" in options:
        to_fetch = glob.glob(root_dir + "/*/STATUTE-" + str(int(options["volume"])))
    elif "volumes" in options:
        start, end = options["volumes"].split("-")
        to_fetch = []
        for v in range(int(start), int(end) + 1):
            to_fetch.extend(glob.glob(root_dir + "/*/STATUTE-" + str(v)))
    elif "year" in options:
        to_fetch = glob.glob(root_dir + "/" + str(int(options["year"])) + "/STATUTE-*")
    elif "years" in options:
        start, end = options["years"].split("-")
        to_fetch = []
        for y in range(int(start), int(end) + 1):
            to_fetch.extend(glob.glob(root_dir + "/" + str(y) + "/STATUTE-*"))
    else:
        to_fetch = sorted(glob.glob(root_dir + "/*/STATUTE-*"))

    logging.warning("Going to process %i volumes" % len(to_fetch))

    utils.process_set(to_fetch, proc_statute_volume, options)


def proc_statute_volume(path, options):
    mods = etree.parse(path + "/mods.xml")
    mods_ns = {"mods": "http://www.loc.gov/mods/v3"}

    # Load the THOMAS committee names for this Congress, which is our best
    # bet for normalizing committee names in the GPO data.
    congress = mods.find("/mods:extension[2]/mods:congress", mods_ns).text
    utils.fetch_committee_names(congress, options)

    logging.warning(f"Processing {path} (Congress {congress})")

    package_id = mods.find("/mods:extension[2]/mods:accessId", mods_ns).text

    for bill in mods.findall("/mods:relatedItem", mods_ns):
        # MODS files also contain information about:
        # ['BACKMATTER', 'FRONTMATTER', 'CONSTAMEND', 'PROCLAMATION', 'REORGPLAN']
        if bill.find("mods:extension/mods:granuleClass", mods_ns).text not in [
            "PUBLICLAW",
            "PRIVATELAW",
            "HCONRES",
            "SCONRES",
        ]:
            continue

        # Get the title and source URL (used in error messages).
        title_text = bill.find("mods:titleInfo/mods:title", mods_ns).text.replace(
            '""', '"'
        )
        source_url = bill.find(
            "mods:location/mods:url[@displayLabel='Content Detail']", mods_ns
        ).text

        # Bill number
        bill_elements = bill.findall(
            "mods:extension/mods:bill[@priority='primary']", mods_ns
        )
        if len(bill_elements) == 0:
            logging.error(
                f"No bill number identified for '{title_text}' ({source_url})"
            )
            continue
        elif len(bill_elements) > 1:
            logging.error("Multiple bill numbers identified for '%s'" % title_text)
            for be in bill_elements:
                logging.error("  -- " + etree.tostring(be).strip())
            logging.error("  @ " + source_url)
            continue
        else:
            bill_congress = bill_elements[0].attrib["congress"]
            bill_type = bill_elements[0].attrib["type"].lower()
            bill_number = bill_elements[0].attrib["number"]
            bill_id = f"{bill_type}{bill_number}-{bill_congress}"

        # Title
        titles = []
        titles.append(
            {
                "title": title_text,
                "as": "enacted",
                "type": "official",
                "is_for_portion": False,
            }
        )

        # Subject
        descriptor = bill.find("mods:extension/mods:descriptor", mods_ns)
        if descriptor is not None:
            subject = descriptor.text
        else:
            subject = None

        # Committees
        committees = []
        cong_committee = bill.find("mods:extension/mods:congCommittee", mods_ns)
        if cong_committee is not None:
            chambers = {"H": "House", "S": "Senate", "J": "Joint"}
            committee = (
                chambers[cong_committee.attrib["chamber"]]
                + " "
                + cong_committee.find("mods:name", mods_ns).text
            )
            committee_info = {
                "committee": committee,
                "activity": [],  # XXX
                "committee_id": utils.committee_names[committee]
                if committee in utils.committee_names
                else None,
            }
            committees.append(committee_info)

        # The 'granuleDate' is the enactment date?
        granule_date = bill.find("mods:extension/mods:granuleDate", mods_ns).text

        sources = [
            {
                "source": "statutes",
                "package_id": package_id,
                "access_id": bill.find("mods:extension/mods:accessId", mods_ns).text,
                "source_url": source_url,
                "volume": bill.find("mods:extension/mods:volume", mods_ns).text,
                "page": bill.find(
                    "mods:part[@type='article']/mods:extent[@unit='pages']/mods:start",
                    mods_ns,
                ).text,
                "position": bill.find("mods:extension/mods:pagePosition", mods_ns).text,
            }
        ]

        law_elements = bill.findall("mods:extension/mods:law", mods_ns)

        # XXX: If <law> is missing, this assumes it is a concurrent resolution.
        #      This may be a problem if the code is updated to accept joint resolutions for constitutional amendments.
        if (law_elements is None) or (len(law_elements) != 1):
            other_chamber = {"HOUSE": "s", "SENATE": "h"}

            actions = [
                {
                    "type": "vote",
                    "vote_type": "vote2",
                    "where": other_chamber[
                        bill.find("mods:extension/mods:originChamber", mods_ns).text
                    ],
                    "result": "pass",  # XXX
                    "how": "unknown",  # XXX
                    #        "text": "",
                    "acted_at": granule_date,  # XXX
                    "status": "PASSED:CONCURRENTRES",
                    "references": [],  # XXX
                }
            ]
        else:
            law_congress = law_elements[0].attrib["congress"]
            law_number = law_elements[0].attrib["number"]
            law_type = (
                "private"
                if (law_elements[0].attrib["isPrivate"] == "true")
                else "public"
            )

            # Check for typos in the metadata.
            if law_congress != bill_congress:
                logging.error(
                    f"Congress mismatch for {bill_type}{bill_number}: {bill_congress} or {law_congress}? ({source_url})"
                )
                continue

            actions = [
                {
                    "congress": law_congress,
                    "number": law_number,
                    "type": "enacted",
                    "law": law_type,
                    "text": f"Became {law_type.capitalize()} Law No: {law_congress}-{law_number}.",
                    "acted_at": granule_date,  # XXX
                    "status": "ENACTED:SIGNED",  # XXX: Check for overridden vetoes!
                    "references": [],  # XXX
                }
            ]

        status, status_date = bill_info.latest_status(actions)

        bill_data = {
            "bill_id": bill_id,
            "bill_type": bill_type,
            "number": bill_number,
            "congress": bill_congress,
            "introduced_at": None,  # XXX
            "sponsor": None,  # XXX
            "cosponsors": [],  # XXX
            "actions": actions,  # XXX
            "history": bill_info.history_from_actions(actions),
            "status": status,
            "status_at": status_date,
            "enacted_as": bill_info.slip_law_from(actions),
            "titles": titles,
            "official_title": bill_info.current_title_for(titles, "official"),
            "short_title": bill_info.current_title_for(titles, "short"),  # XXX
            "popular_title": bill_info.current_title_for(titles, "popular"),  # XXX
            "subjects_top_term": subject,
            "subjects": [],
            "related_bills": [],  # XXX: <associatedBills> usually only lists the current bill.
            "committees": committees,
            "amendments": [],  # XXX
            "sources": sources,
            "updated_at": datetime.datetime.fromtimestamp(time.time()),
        }

        if not options.get("textversions", False):
            bill_info.output_bill(bill_data, options)

        # XXX: Can't use bill_versions.fetch_version() because it depends on fdsys.
        version_code = "enr"
        bill_version_id = f"{bill_type}{bill_number}-{bill_congress}-{version_code}"
        bill_version = {
            "bill_version_id": bill_version_id,
            "version_code": version_code,
            "issued_on": status_date,
            "urls": {
                "pdf": bill.find(
                    "mods:location/mods:url[@displayLabel='PDF rendition']", mods_ns
                ).text
            },
            "sources": sources,
        }
        utils.write(
            json.dumps(
                bill_version, sort_keys=True, indent=2, default=utils.format_datetime
            ),
            govinfo.output_for_bill_version(bill_version_id),
        )

        # Process the granule PDF.
        # - Hard-link it into the right place to be seen as bill text.
        # - Run "pdftotext -layout" to convert it to plain text and save it in the bill text location.
        pdf_file = path + "/" + sources[0]["access_id"] + "/document.pdf"
        if os.path.exists(pdf_file):
            dst_path = bills.output_for_bill(
                bill_data["bill_id"], "text-versions/" + version_code, is_data_dot=False
            )
            if options.get("linkpdf", False):
                os.link(pdf_file, dst_path + "/document.pdf")  # a good idea
            if options.get("extracttext", False):
                logging.error("Running pdftotext on %s..." % pdf_file)
                if (
                    subprocess.call(
                        ["pdftotext", "-layout", pdf_file, dst_path + "/document.txt"]
                    )
                    != 0
                ):
                    raise Exception("pdftotext failed on %s" % pdf_file)

    return {"ok": True, "saved": True}



"""
A module that monkey-patches the output_bill method to push the bill identifier
onto a task queue after the data file has been written to disk. To use this
module, invoke the bills scraper with the --patch option like so:

  ./run bills --patch=contrib.beanstalkd

You must include a 'beakstalk' section in config.yml with this structure
(though the values are up to you):

  beanstalk:
    connection:
      host: 'localhost'
      port: 11300
    tubes:
      bills: 'us_bills'
      amendments: 'us_amendments'
      votes: 'us_votes'
"""


import logging
import sys
import traceback
from collections import Counter
from functools import wraps

import beanstalkc
import yaml

# The patch module is loaded after the task module is loaded, so all task
# modules are on the import path.
from congress_crawler import amendment_info, bills, vote_info


__all__ = [
    "patch",
    "process_bill_wrapper",
    "process_amendment_wrapper",
    "output_vote_wrapper",
]


_Connection = None
_Config = None


def init_guard(reconnect=False):
    global _Config, _Connection
    if _Config is None:
        with open("config.yml") as conffile:
            config = yaml.load(conffile)
            assert "beanstalk" in config
            assert "connection" in config["beanstalk"]
            assert "host" in config["beanstalk"]["connection"]
            assert "port" in config["beanstalk"]["connection"]
            assert "tubes" in config["beanstalk"]
            assert "bills" in config["beanstalk"]["tubes"]
            assert "amendments" in config["beanstalk"]["tubes"]
            assert "votes" in config["beanstalk"]["tubes"]
            tube_names = list(config["beanstalk"]["tubes"].values())
            assert (
                max(Counter(tube_names).values()) == 1
            ), "Must use unique beanstalk tube names."
            _Config = config["beanstalk"]
    if _Connection is None or reconnect is True:
        conn = beanstalkc.Connection(**_Config["connection"])
        assert conn is not None
        _Connection = conn
    return _Connection, _Config


def process_bill_wrapper(process_bill):
    @wraps(process_bill)
    def _process_bill(bill, options, *args, **kwargs):
        orig_result = process_bill(bill, options, *args, **kwargs)

        (conn, config) = init_guard()
        for _ in range(2):
            try:
                conn.use(config["tubes"]["bills"])
                conn.put(bill)
                logging.warning(f"Queued {bill} to beanstalkd.")
                break
            except beanstalkc.SocketError:
                logging.warning(
                    "Lost connection to beanstalkd. Attempting to reconnect."
                )
                (conn, config) = init_guard(reconnect=True)
            except Exception as e:
                logging.warning(
                    f"Ignored exception while queueing bill to beanstalkd: {str(type(e))} {str(e)}"
                )
                traceback.print_exc()
                break

        return orig_result

    return _process_bill


def process_amendment_wrapper(process_amendment):
    @wraps(process_amendment)
    def _process_amendment(amdt_dict, bill_id, options, *args, **kwargs):
        orig_result = process_amendment(amdt_dict, bill_id, options, *args, **kwargs)
        amdt = amendment_info.build_amendment_id(
            amdt_dict["type"].lower(), amdt_dict["number"], amdt_dict["congress"]
        )

        (conn, config) = init_guard()
        for _ in range(2):
            try:
                conn.use(config["tubes"]["amendments"])
                conn.put(str(amdt))
                logging.warning(f"Queued {amdt} to beanstalkd.")
                break
            except beanstalkc.SocketError:
                logging.warning(
                    "Lost connection to beanstalkd. Attempting to reconnect."
                )
                (conn, config) = init_guard(reconnect=True)
            except Exception as e:
                logging.warning(
                    f"Ignored exception while queueing amendment to beanstalkd: {str(type(e))} {str(e)}"
                )
                traceback.print_exc()
                break

        return orig_result

    return _process_amendment


def output_vote_wrapper(output_vote):
    @wraps(output_vote)
    def _output_vote(vote, options, *args, **kwargs):
        orig_result = output_vote(vote, options, *args, **kwargs)

        (conn, config) = init_guard()
        for _ in range(2):
            try:
                conn.use(config["tubes"]["votes"])
                conn.put(vote["vote_id"])
                logging.warning("Queued {} to beanstalkd.".format(vote["vote_id"]))
                break
            except beanstalkc.SocketError:
                logging.warning(
                    "Lost connection to beanstalkd. Attempting to reconnect."
                )
                (conn, config) = init_guard(reconnect=True)
            except Exception as e:
                logging.warning(
                    f"Ignored exception while queueing vote to beanstalkd: {str(type(e))} {str(e)}"
                )
                traceback.print_exc()
                break

        return orig_result

    return _output_vote


def patch(task_name):
    bills.process_bill = process_bill_wrapper(bills.process_bill)
    amendment_info.process_amendment = process_amendment_wrapper(
        amendment_info.process_amendment
    )
    vote_info.output_vote = output_vote_wrapper(vote_info.output_vote)


# Avoid scraping if the beanstalk config is invalid.
try:
    init_guard()
except AssertionError:
    print(__doc__)
    sys.exit(1)



#!/usr/bin/env python

from pathlib import Path

from setuptools import find_packages, setup

from congress_crawler import __version__


with (Path(__file__).parent / "requirements.txt").open() as f:
    required = f.read().splitlines()


setup(
    name="congress-crawler",
    version=__version__,
    author="Andrew Chen Wang",
    author_email="acwangpython@gmail.com",
    url="https://github.com/Hear-Ye/congress",
    description="Gathers data on the U.S. Congress.",
    long_description=open("README.md").read(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
    ],
    license="CC0",
    packages=find_packages(),
    install_requires=required,
    include_package_data=True,
    zip_safe=False,
)



import json
import logging
import os
import re
import subprocess
import time
from datetime import datetime

import lxml
from bs4 import BeautifulSoup
from dateutil.relativedelta import MO, relativedelta

from congress_crawler import utils
from congress_crawler.bills import output_for_bill


# Parsing data from the House' upcoming floor feed, at
# https://docs.house.gov/floor/
#
# This contains data on what bills and draft bills are coming up
# on the floor of the House.
#
# This script will transform the data in the provided XML feed to JSON,
# and download associated documents to disk.
#
# TODO:
#   * Detect and extract any XML files attached to PDFs.
#   * parsing out metadata from any provided XML documents.
#   * handle 'subitems' (e.g. House committee reports)
#
# options:
#   week_of: the date of a Monday of a week to look for. defaults to current week.
#   download: download associated documents and convert PDFs to text


def run(options):
    # accepts yyyymmdd format
    given_week = options.get("week_of", None)
    if given_week is None:
        for_the_weeks = get_mondays_to_scan(options)
    else:
        for_the_weeks = [get_monday_of_week(given_week)]

    # fetch info
    for for_the_week in for_the_weeks:
        run_for_week(for_the_week, options)


def run_for_week(for_the_week, options):
    logging.info(
        "Scraping upcoming bills from docs.house.gov/floor for the week of %s..."
        % for_the_week
    )
    house_floor = fetch_floor_week(for_the_week, options)
    if house_floor is None:
        logging.warning("Nothing posted for the week of %s" % for_the_week)
        return

    output_file = f"{utils.data_dir()}/upcoming_house_floor/{for_the_week}.json"
    output = json.dumps(
        house_floor, sort_keys=True, indent=2, default=utils.format_datetime
    )
    utils.write(output, output_file)

    logging.warning(
        "Found %i bills for the week of %s, written to %s"
        % (len(house_floor["upcoming"]), for_the_week, output_file)
    )


# For any week, e.g. https://docs.house.gov/floor/Download.aspx?file=/billsthisweek/20131021/20131021.xml
def fetch_floor_week(for_the_week, options):
    base_url = "https://docs.house.gov/floor/Download.aspx?file=/billsthisweek/"
    week_url = base_url + f"{for_the_week}/{for_the_week}.xml"

    # Turn on 'force' to re-download the schedules, by default, since the content
    # changes frequently and we're scanning weeks that might have 404'd previously
    # when we looked ahead. We leave 'force' off for downloading the file attachments.
    options2 = dict(options)
    if "force" not in options2:
        options2["force"] = True

    body = utils.download(
        week_url, "upcoming_house_floor/%s.xml" % for_the_week, options2
    )
    if "was not found" in body:
        return None
    dom = lxml.etree.fromstring(body)

    # can download the actual attached files to disk, if asked
    download = options.get("download", False)

    # always present at the feed level
    congress = int(dom.xpath("//floorschedule")[0].get("congress-num"))

    # week of this day, e.g. '2013-01-21'
    legislative_day = (
        for_the_week[0:4] + "-" + for_the_week[4:6] + "-" + for_the_week[6:]
    )

    upcoming = []

    for node in dom.xpath("//floorschedule/category/floor-items/floor-item"):
        bill_number = node.xpath("legis-num//text()")[0]

        # TODO: fetch non-bills too
        if not bill_number:
            logging.warning("Skipping item, not a bill: %s" % description)
            continue

        description = node.xpath("floor-text//text()")[0]

        # how is this bill being considered?
        category = next(node.iterancestors("category")).get("type")
        if "suspension" in category:
            consideration = "suspension"
        elif "pursuant" in category:
            consideration = "rule"
        else:
            consideration = "unknown"

        logging.warning("[%s]" % bill_number)

        # todo: establish most recent date from a combo of added, published, updates
        date = date_for(node.get("publish-date"))

        # all items will have this
        bill = {
            "description": description,
            "floor_item_id": node.get("id"),
            "consideration": consideration,
            "published_at": date_for(node.get("publish-date")),
            "added_at": date_for(node.get("add-date")),
        }

        # treat drafts and numbered bills a bit differently
        if "_" in bill_number:
            draft_bill_id = draft_bill_id_for(bill_number, date, congress)
            bill["item_type"] = "draft_bill"
            bill["draft_bill_id"] = draft_bill_id
        else:
            m = re.match(
                "(Concur(ring)? in )?(?P<type>((the )?(Senate|House) Amendments? (with an amendment )?to )+)(?P<bill>.*)",
                bill_number,
                re.I,
            )
            if m:
                amendment_type = m.group("type").split("to")[0]
                if "Senate" in amendment_type and "House" not in amendment_type:
                    bill["item_type"] = "senate_amendment"
                elif "House" in amendment_type and "Senate" not in amendment_type:
                    bill["item_type"] = "house_amendment"
                else:
                    raise ValueError(bill_number)
                bill_number = m.group("bill")

            elif re.match("Conference report to accompany ", bill_number, re.I):
                bill["item_type"] = "conference_report"
                bill_number = bill_number.lower().replace(
                    "conference report to accompany ", ""
                )
            else:
                bill["item_type"] = "bill"

            # In one case we got "H. Res. 497 (H. Rept. 116-125)".
            # Stop at parens.
            bill_number = re.sub(r"\(.*", "", bill_number)

            try:
                bill["bill_id"] = bill_id_for(bill_number.strip(), congress)
            except ValueError:
                logging.error("Could not parse bill from: %s" % bill_number)
                continue

        bill["files"] = []
        for file in node.xpath("files/file"):
            file_url = file.get("doc-url")
            filename = file_url.split("/")[-1]
            file_format = file.get("doc-type").lower()

            logging.warning(
                f"\t{file_format.upper()} file for {bill_number}: {filename}"
            )

            file_field = {
                "url": file_url,
                "format": file_format,
                "added_at": date_for(file.get("add-date")),
                "published_at": date_for(file.get("publish-date")),
            }

            bill["files"].append(file_field)

            # now try downloading the file to disk and linking it to the data
            if not download:
                continue
            try:
                file_path = f"upcoming_house_floor/{for_the_week}/{filename}"
                try:
                    os.makedirs(
                        os.path.join(utils.data_dir(), os.path.dirname(file_path))
                    )
                except OSError:
                    pass  # directory exists
                options3 = dict(options)
                options3["to_cache"] = False  # put in the actual specified directory
                options3["binary"] = True  # force binary mode, no file escaping
                utils.download(
                    file_url, os.path.join(utils.data_dir(), file_path), options3
                )
                file_field["path"] = file_path
            except OSError:
                logging.error(
                    "Omitting 'path', couldn't download file {} from House floor for the week of {}".format(
                        file_field["url"], for_the_week
                    )
                )
                continue

            # if it's a PDF, convert to text and extract XML
            if file_format == "pdf" and file_path.endswith(".pdf"):
                # extract text
                text_path = file_path.replace(".pdf", ".txt")
                if (
                    subprocess.call(
                        [
                            "pdftotext",
                            "-layout",
                            os.path.join(utils.data_dir(), file_path),
                            os.path.join(utils.data_dir(), text_path),
                        ],
                        universal_newlines=True,
                    )
                    != 0
                ):
                    raise Exception("pdftotext failed on %s" % file_path)
                file_field["text_path"] = text_path

                # extract embedded XML
                for line in subprocess.check_output(
                    ["pdfdetach", "-list", os.path.join(utils.data_dir(), file_path)],
                    universal_newlines=True,
                ).split("\n"):
                    m = re.match(r"(\d+):\s*(.*)", line)
                    if m:
                        attachment_n, attachment_fn = m.groups()
                        if attachment_fn.endswith(".xml"):
                            text_path = file_path.replace(".pdf", ".xml")
                            subprocess.check_call(
                                [
                                    "pdfdetach",
                                    os.path.join(utils.data_dir(), file_path),
                                    "-save",
                                    attachment_n,
                                    "-o",
                                    os.path.join(utils.data_dir(), text_path),
                                ],
                                universal_newlines=True,
                            )
                            file_field["xml_path"] = text_path

        upcoming.append(bill)

        if "bill_id" in bill:
            # Save this bill data to the bill's bill text directory.
            text_data_path = output_for_bill(
                bill["bill_id"],
                os.path.join("text-versions", "dhg-" + bill["floor_item_id"] + ".json"),
                is_data_dot=False,
            )
            try:
                os.makedirs(
                    os.path.join(utils.data_dir(), os.path.dirname(text_data_path))
                )
            except OSError:
                pass  # directory exists
            utils.write(
                json.dumps(
                    bill, sort_keys=True, indent=2, default=utils.format_datetime
                ),
                text_data_path,
            )

    # Create and return the house floor file data.
    house_floor = {
        "congress": congress,
        "week_of": legislative_day,
        "upcoming": upcoming,
    }

    return house_floor


def get_monday_of_week(day_to_get_bills):
    formatted_day = datetime.strptime(day_to_get_bills, "%Y%m%d").date()
    return (formatted_day + relativedelta(weekday=MO(-1))).strftime("%Y%m%d")


# actually go fetch docs.house.gov/floor/ and scrape the download link out of it


def get_latest_monday(options):
    # docs.house.gov always links to the most recent week that isn't in the future.
    url = "https://docs.house.gov/floor/"
    html = utils.download(url, None, options)
    doc = BeautifulSoup(html, features="lxml")

    links = doc.select("a.downloadXML")
    if len(links) != 1:
        utils.admin(
            "There is no docs.house.gov download link --- maybe there are no upcoming bills."
        )
        return None

    link = links[0]
    week = os.path.split(link["href"])[-1].split(".")[0]
    week = datetime.strptime(week, "%Y%m%d").date()

    return week


def get_mondays_to_scan(options):
    # Get the week currently linked on docs.house.gov. If there isn't any (e.g. we are between
    # sessions), just return an empty list of weeks to scan.
    most_recent = get_latest_monday(options)
    if most_recent is None:
        return []

    # Look two weeks into the future too, since when we get to the end of the week the next
    # week's list is sometimes available, and sometimes a week beyond that.
    return [
        (most_recent + relativedelta(days=7 * i)).strftime("%Y%m%d") for i in [0, 1, 2]
    ]


def bill_id_for(bill_number, congress):
    number = bill_number.replace(".", "").replace(" ", "").lower()
    if not re.match(r"^(hr|s|hres|sres|hjres|sjres|hconres|sconres)\d{1,4}$", number):
        raise ValueError(number)
    return "%s-%i" % (number, congress)


def draft_bill_id_for(bill_number, published_at, congress):
    number = bill_number.replace(".", "").replace(" ", "").replace("_", "").lower()
    epoch = time.mktime(published_at.timetuple())
    return "%s%i-%i" % (number, epoch, congress)


def date_for(timestamp):
    if "." not in timestamp:
        strptime_config = "%Y-%m-%dT%H:%M:%S"
    else:
        strptime_config = "%Y-%m-%dT%H:%M:%S.%f"
    return datetime.strptime(timestamp, strptime_config)



import datetime
import json
import logging
import os
import os.path
import re
import urllib.parse

from dateutil.parser import isoparse
from lxml import etree, html

from congress_crawler import utils, vote_info


def run(options):
    vote_id = options.get("vote_id", None)

    if vote_id:
        vote_chamber, vote_number, congress, session_year = utils.split_vote_id(vote_id)
        to_fetch = [vote_id]
    else:
        congress = options.get("congress", None)
        if congress:
            session_year = options.get("session", None)
            if not session_year:
                logging.error("If you provide a --congress, provide a --session year.")
                return None
        else:
            congress = utils.current_congress()
            session_year = options.get("session", str(utils.current_legislative_year()))

        chamber = options.get("chamber", None)

        if chamber == "house":
            to_fetch = vote_ids_for_house(congress, session_year, options)
        elif chamber == "senate":
            to_fetch = vote_ids_for_senate(congress, session_year, options)
        else:
            to_fetch = (vote_ids_for_house(congress, session_year, options) or []) + (
                vote_ids_for_senate(congress, session_year, options) or []
            )

        if not to_fetch:
            if not options.get("fast", False):
                logging.error("Error figuring out which votes to download, aborting.")
            else:
                logging.warning("No new or recent votes.")
            return None

        limit = options.get("limit", None)
        if limit:
            to_fetch = to_fetch[: int(limit)]

    if options.get("pages_only", False):
        return None

    logging.warning(
        f"Going to fetch {len(to_fetch)} votes from congress "
        f"#{congress} session {session_year}"
    )

    utils.process_set(to_fetch, vote_info.fetch_vote, options)


# page through listing of House votes of a particular congress and session


def vote_ids_for_house(congress, session_year, options):
    vote_ids = []

    index_page = f"https://clerk.house.gov/evs/{session_year}/index.asp"
    group_page = r"ROLL_(\d+)\.asp"
    link_pattern = (
        r"http://clerk.house.gov/cgi-bin/vote.asp\?year=%s&rollnumber=(\d+)"
        % session_year
    )

    # download index page, find the matching links to the paged listing of votes
    page = utils.download(
        index_page, f"{congress}/votes/{session_year}/pages/house.html", options
    )

    if not page:
        logging.error("Couldn't download House vote index page, skipping")
        return None

    # extract matching links
    doc = html.document_fromstring(page)
    links = doc.xpath(
        f"//a[re:match(@href, '{group_page}')]",
        namespaces={"re": "http://exslt.org/regular-expressions"},
    )

    for link in links:
        # get some identifier for this inside page for caching
        grp = re.match(group_page, link.get("href")).group(1)

        # download inside page, find the matching links
        page = utils.download(
            urllib.parse.urljoin(index_page, link.get("href")),
            f"{congress}/votes/{session_year}/pages/house_{grp}.html",
            options,
        )

        if not page:
            logging.error(f"Couldn't download House vote group page ({grp}), aborting")
            continue

        doc = html.document_fromstring(page)
        votelinks = doc.xpath(
            f"//a[re:match(@href, '{link_pattern}')]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

        for votelink in votelinks:
            num = re.match(link_pattern, votelink.get("href")).group(1)
            vote_id = f"h{num}-{congress}.{session_year}"
            if not should_process(vote_id, options):
                continue
            vote_ids.append(vote_id)

    return utils.uniq(vote_ids)


def vote_ids_for_senate(congress, session_year, options):
    session_num = int(session_year) - utils.get_congress_first_year(int(congress)) + 1

    vote_ids = []

    url = (
        "https://www.senate.gov/legislative/LIS/roll_call_lists/vote_menu_%s_%d.xml"
        % (congress, session_num)
    )
    page = utils.download(
        url,
        f"{congress}/votes/{session_year}/pages/senate.xml",
        utils.merge(options, {"binary": True}),
    )

    if not page or b"Requested Page Not Found (404)" in page:
        logging.error(f"Couldn't download Senate vote XML index {url}, skipping")
        return None

    dom = etree.fromstring(page)

    # Sanity checks.
    if int(congress) != int(dom.xpath("congress")[0].text):
        logging.error(
            f"Senate vote XML returns the wrong Congress: "
            f"{dom.xpath('congress')[0].text}"
        )
        return None
    if int(session_year) != int(dom.xpath("congress_year")[0].text):
        logging.error(
            "Senate vote XML returns the wrong session: "
            f"{dom.xpath('congress_year')[0].text}"
        )
        return None

    # Get vote list.
    for vote in dom.xpath("//vote"):
        num = int(vote.xpath("vote_number")[0].text)
        vote_id = f"s{num}-{congress}.{session_year}"
        if not should_process(vote_id, options):
            continue
        vote_ids.append(vote_id)
    return vote_ids


def should_process(vote_id, options):
    if not options.get("fast", False):
        return True

    # If --fast is used, only download new votes or votes taken in the last
    # three days (when most vote changes and corrections should occur).
    f = vote_info.output_for_vote(vote_id, "json")
    if not os.path.exists(f):
        return True

    v = json.load(open(f))
    now = utils.eastern_time_zone.localize(datetime.datetime.now())
    return (now - isoparse(v["date"])) < datetime.timedelta(days=3)
